/*
	GnuJSP - a free JSP implementation
	Copyright (C) 1998, Vincent Partington <vinny@nmg.nl>
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package	nl.nmg.jsp;
import	java.io.BufferedReader;
import	java.io.ByteArrayOutputStream;
import	java.io.File;
import	java.io.FileNotFoundException;
import	java.io.FileReader;
import	java.io.FileWriter;
import	java.io.IOException;
import	java.io.PrintStream;
import	java.io.PrintWriter;
import	java.util.Enumeration;
import	java.util.StringTokenizer;
import	java.util.Vector;

public class JSPCompiler {
	public static final int		COMPILER_VERSION_NR = 1;

	private static final int	PARSE_NONE = 0,
								PARSE_HTML = 1,
								PARSE_DECLARATION = 2,
								PARSE_SCRIPTLET = 3,
								PARSE_EXPRESSION = 4;

	private static final String	lineSeparator = System.getProperty("line.separator"),
								encodedLineSeparator = cEncode(System.getProperty("line.separator")),
								classPath = System.getProperty("java.class.path");

	private File				classFilesDir,
								jspFile,
								javaFile;
	private String				className;

	private int					parseState = PARSE_NONE;
	private String				key, value;

	private Vector				dependencies = new Vector();
	private String				content_typeDirective = null,
								extendsDirective = null,
								languageDirective = null,
								methodDirective = null;
	private StringBuffer		implementsDirective = new StringBuffer(),
								importsDirective = new StringBuffer(),
								declarationsCode = new StringBuffer(),
								scriptletsCode = new StringBuffer();

	/**
		@params classFilesDir the directory where the java and class file should be written
		@params jspFile the file containing the JSP code to compile
		@params className the classname of the file
	*/
	public JSPCompiler(File classFilesDir, File jspFile, String className) {
		this.classFilesDir = classFilesDir;
		this.jspFile = jspFile;
		this.javaFile = getFileForClass(classFilesDir, className, ".java");
		this.className = className;
	}

	public void compile() throws JSPException {
		parseJspFile();
		generateJavaFile();
		compileJavaFile();
	}

	private void parseJspFile() throws JSPException {
		parseOneJspFile(jspFile);
		changeParseState(PARSE_NONE);
	
		if(content_typeDirective == null) {
			content_typeDirective = "text/html";
		}
		if(extendsDirective == null) {
			extendsDirective = "javax.servlet.http.HttpServlet";
		}
		if(languageDirective != null && !languageDirective.equals("java")) {
			throw new JSPException("The only supported language is java");
		}
		if(methodDirective == null) {
			methodDirective = "service";
		}
	}

	private void parseOneJspFile(File jspFile) throws JSPException {
		BufferedReader	in;
		StringBuffer	srcBuf;
		String			s, src;
		StringTokenizer	toker;
		int				i, srcPos, srcLen;
		char			c, d;

		// read jsp source
		dependencies.addElement(jspFile);
		dependencies.addElement(new Long(jspFile.lastModified()));
		try {
			in = new BufferedReader(new FileReader(jspFile));
			try {
				srcBuf = new StringBuffer((int) jspFile.length());
				while((s = in.readLine()) != null) {
					srcBuf.append(s).append((char) '\n');
				}
				src = srcBuf.toString();
				srcBuf = null;
			} finally {
				in.close();
			}
		} catch(FileNotFoundException fnfexc) {
			throw new JSPException("File " + jspFile + " not found");
		} catch(IOException ioexc) {
			throw new JSPException(ioexc.toString());
		}

		// parse jsp source
		srcLen = src.length();
		srcPos = 0;
		PARSING: while(srcPos < srcLen) {
			c = src.charAt(srcPos);

			// check state change
			if(c == '<' || c == '%') {
				if(parseState == PARSE_NONE || parseState == PARSE_HTML) {
					if(c == '<' && srcPos < srcLen-3 && src.charAt(srcPos+1) == '%') {
						d = src.charAt(srcPos+2);
						if(d == '=') {
							// start of expression
							changeParseState(PARSE_EXPRESSION);
							srcPos += 3;
							continue PARSING;
						} else if(d == '@') {
							// directive
							i = parseKeyAndValue(src, srcPos+3, srcLen);
							for(;;) {
								if(i == -1) {
									continue PARSING;
								}
								srcPos = i;

								if(key.equals("content_type") && content_typeDirective == null) {
									content_typeDirective = value;
								} else if(key.equals("extends") && extendsDirective == null) {
									extendsDirective = value;
								} else if(key.equals("implements")) {
									toker = new StringTokenizer(value, " ,");
									while(toker.hasMoreTokens()) {
										if(implementsDirective.length() == 0) {
											implementsDirective.append(" implements ");
										} else {
											implementsDirective.append(", ");
										}
										implementsDirective.append(toker.nextToken());
									}
								} else if(key.equals("import")) {
									toker = new StringTokenizer(value, " ,");
									while(toker.hasMoreTokens()) {
										importsDirective.append("import ").append(toker.nextToken()).append((char) ';').append(lineSeparator);
									}
								} else if(key.equals("include")) {
									parseOneJspFile(new File(jspFile.getParent(), value));
								} else if(key.equals("language") && languageDirective == null) {
									languageDirective = value;
								} else if(key.equals("method") && methodDirective == null) {
									methodDirective = value;
								}

								if(i >= srcLen - 1) {
									continue PARSING;
								} else if(src.charAt(i) == '%' && src.charAt(i+1) == '>') {
									srcPos = i + 2;
									continue PARSING;
								}
								i = parseKeyAndValue(src, srcPos, srcLen);
							}
						} else {
							// start of scriplet
							changeParseState(PARSE_SCRIPTLET);
							srcPos += 2;
							continue PARSING;
						}
					} else if(c == '<' && srcPos < srcLen-20 &&
								src.substring(srcPos+1, srcPos+7).toUpperCase().equals("SCRIPT") &&
								Character.isWhitespace(src.charAt(srcPos+7)))
					{
						// possible start of declarations
						boolean runatServer = false;
						i = srcPos + 8;
						while(i < srcLen - 3) {
							if(src.charAt(i) == '>') {
								if(runatServer) {
									// start of declaration
									changeParseState(PARSE_DECLARATION);
									srcPos = i+1;
									continue PARSING;
								} else {
									break;
								}
							}
							i = parseKeyAndValue(src, i, srcLen);
							if(i == -1) {
								break;
							} else if(key.toUpperCase().equals("RUNAT") && value.toUpperCase().equals("SERVER")) {
								runatServer = true;
							}
						}
					}
				} else if(parseState == PARSE_EXPRESSION || parseState == PARSE_SCRIPTLET) {
					if(c == '%' && srcPos < srcLen-1 && src.charAt(srcPos+1) == '>') {
						// end of expression or scriptlet
						changeParseState(PARSE_NONE);
						srcPos += 2;
						continue PARSING;
					}
				} else if(parseState == PARSE_DECLARATION) {
					if(c == '<' && srcPos < srcLen-8 &&
							src.substring(srcPos+1, srcPos+8).toUpperCase().equals("/SCRIPT") &&
							(Character.isWhitespace(src.charAt(srcPos+8)) || src.charAt(srcPos+8) == '>'))
					{
						// end of declaration
						changeParseState(PARSE_NONE);
						while(srcPos < srcLen) {
							if(src.charAt(srcPos++) == '>') {
								break;
							}
						}
						continue PARSING;
					}
				}
			}

			// output character
			switch(parseState) {
			case PARSE_NONE:
				changeParseState(PARSE_HTML);
			case PARSE_HTML:
				switch(c) {
				case '\b':
					scriptletsCode.append("\\b");
					break;
				case '\t':
					scriptletsCode.append("\\t");
					break;
				case '\n':
					scriptletsCode.append(encodedLineSeparator).append("\" +").append(lineSeparator).append("\t\t\t\"");
					break;
				case '\f':
					scriptletsCode.append("\\f");
					break;
				case '\'':
					scriptletsCode.append("\\\'");
					break;
				case '\"':
					scriptletsCode.append("\\\"");
					break;
				default:
					if(c > 0xFF) {
						s = "00" + Integer.toHexString(c);
						scriptletsCode.append("\\u").append(s.substring(s.length() - 4));
					} else if(c < ' ' || c >= 0x7F) {
						s = "00" + Integer.toOctalString(c);
						scriptletsCode.append((char) '\\').append(s.substring(s.length() - 3));
					} else {
						scriptletsCode.append((char) c);
					}
					break;
				}
				break;
			case PARSE_DECLARATION:
				if(c == '\n')
					declarationsCode.append(lineSeparator);
				else
					declarationsCode.append((char) c);
				break;
			default:
				if(c == '\n')
					scriptletsCode.append(lineSeparator);
				else
					scriptletsCode.append((char) c);
				break;
			}
			srcPos++;
		}
	}

	void changeParseState(int newState) {
		if(newState == parseState)
			return;

		switch(parseState) {
		case PARSE_HTML:
			scriptletsCode.append("\");").append(lineSeparator);
			break;
		case PARSE_EXPRESSION:
			scriptletsCode.append("));").append(lineSeparator);
			break;
		case PARSE_SCRIPTLET:
			scriptletsCode.append(lineSeparator);
			break;
		}

		parseState = newState;

		switch(parseState) {
		case PARSE_HTML:
			scriptletsCode.append("\t\tout.print(\"");
			break;
		case PARSE_EXPRESSION:
			scriptletsCode.append("\t\tout.print(__valueOf(");
			break;
		}
	}

	private int parseKeyAndValue(String src, int srcPos, int srcLen) {
		char	c;
		int		oldSrcPos;

		// skip whitespace
		while(srcPos < srcLen && Character.isWhitespace(src.charAt(srcPos)))
			srcPos++;
		if(srcPos >= srcLen)
			return -1;

		// gather key
		oldSrcPos = srcPos;
		while(srcPos < srcLen) {
			c = src.charAt(srcPos);
			if(!Character.isLetterOrDigit(c) && c != '-' && c != '_')
				break;
			srcPos++;
		}
		if(srcPos >= srcLen || oldSrcPos == srcPos)
			return -1;
		key = src.substring(oldSrcPos, srcPos).toLowerCase();

		// skip whitespace
		while(srcPos < srcLen && Character.isWhitespace(src.charAt(srcPos)))
			srcPos++;
		if(srcPos >= srcLen)
			return -1;

		// gather optional value
		if(src.charAt(srcPos) == '=') {
			srcPos++;

			// skip whitespace
			while(srcPos < srcLen && Character.isWhitespace(src.charAt(srcPos)))
				srcPos++;
			if(srcPos >= srcLen)
				return -1;

			c = src.charAt(srcPos);
			if(c == '\'' || c == '\"') {
				srcPos++;
				oldSrcPos = srcPos;
				while(srcPos < srcLen && src.charAt(srcPos) != c)
					srcPos++;
				if(srcPos >= srcLen)
					return -1;
				value = src.substring(oldSrcPos, srcPos);
				srcPos++;
			} else {
				oldSrcPos = srcPos;
				while(srcPos < srcLen) {
					c = src.charAt(srcPos);
					if(Character.isWhitespace(c) || c == '%' || c == '>')
						break;
					srcPos++;
				}
				if(srcPos >= srcLen || oldSrcPos == srcPos)
					return -1;
				value = src.substring(oldSrcPos, srcPos);
			}

			// skip whitespace
			while(srcPos < srcLen && Character.isWhitespace(src.charAt(srcPos)))
				srcPos++;
			if(srcPos >= srcLen)
				return -1;
		} else {
			value = key;
		}

		return srcPos;
	}

	private void generateJavaFile() throws JSPException {
		PrintWriter 	out;
		Enumeration		e;
		File			f;
		String			pn, cn;
		Long			l;
		int				i;

		try {
			f = new File(javaFile.getParent());
			f.mkdirs();
			out = new PrintWriter(new FileWriter(javaFile));
			i = className.lastIndexOf('.');
			if(i > 0) {
				pn = className.substring(0, i);
				cn = className.substring(i+1);
			} else {
				pn = null;
				cn = className;
			}
			try {
				// generate header and start of class
				out.println("/* Code generated by nl.nmg.jsp.JSPServlet, do not edit! */");
				if(pn != null) {
					out.println("package " + pn + ";");
				}
				out.println(importsDirective);
				out.println("public final class " + cn +
							" extends " + extendsDirective + implementsDirective);
				out.println("{");

				// generate declarations
				out.println(declarationsCode);

				// generate specified method
				out.println("\tpublic void " + methodDirective +
							"(javax.servlet.http.HttpServletRequest request, " +
							"javax.servlet.http.HttpServletResponse response) " +
							"throws javax.servlet.ServletException, java.io.IOException");
				out.println("\t{");
				out.println("\t\tjava.io.BufferedReader in = request.getReader();");
				out.println("\t\tresponse.setContentType(\"" + cEncode(content_typeDirective) + "\");");
				out.println("\t\tjava.io.PrintWriter out = response.getWriter();");
				out.println("\t\ttry {");
				out.println(scriptletsCode);
				out.println("\t\t} finally {");
				out.println("\t\t\tout.close();");
				// out.println("\t\t\tin.close();");	// werkt niet met JRun
				out.println("\t\t}");
				out.println("\t}");
				out.println();

				// generate own version of String.valueOf()
				out.println("\tprivate static String __valueOf(java.lang.Object obj) {");
				out.println("\t\tif(obj == null) {");
				out.println("\t\t\treturn \"\";");
				out.println("\t\t} else {");
				out.println("\t\t\treturn obj.toString();");
				out.println("\t\t}");
				out.println("\t}");
				out.println("\tprivate static String __valueOf(boolean b) { return String.valueOf(b); }");
				out.println("\tprivate static String __valueOf(char c)    { return String.valueOf(c); }");
				out.println("\tprivate static String __valueOf(int i)     { return String.valueOf(i); }");
				out.println("\tprivate static String __valueOf(long l)    { return String.valueOf(l); }");
				out.println("\tprivate static String __valueOf(float f)   { return String.valueOf(f); }");
				out.println("\tprivate static String __valueOf(double d)  { return String.valueOf(d); }");
				out.println();

				// generate dependency check code
				out.println("\tpublic static int __compilerVersionNr = " + COMPILER_VERSION_NR + ";");
				out.println("\tprivate static java.lang.Object[] __dependencies = new java.lang.Object[]{");
				e = dependencies.elements();
				for(;;) {
					f = (File) e.nextElement();
					l = (Long) e.nextElement();
					out.println("\t\tnew java.io.File(\"" + cEncode(f.getAbsolutePath()) + "\"),");
					out.print("\t\tnew java.lang.Long(" + l + "L)");
					if(e.hasMoreElements()) {
						out.println(",");
					} else {
						out.println();
						break;
					}
				}
				out.println("\t};");
				out.println("\tpublic static boolean __checkDependencies() {");
				out.println("\t\tfor(int i = 0; i < __dependencies.length; i += 2) {");
				out.println("\t\t\tif(!((java.io.File) __dependencies[i]).exists() ||");
				out.println("\t\t\t\t\t((java.io.File) __dependencies[i]).lastModified() > ((java.lang.Long) __dependencies[i+1]).longValue()) {");
				out.println("\t\t\t\treturn true;");
				out.println("\t\t\t}");
				out.println("\t\t}");
				out.println("\t\treturn false;");
				out.println("\t}");

				// generate end of class
				out.println("}");
			} finally {
				out.close();
			}
		} catch(IOException ioexc) {
			throw new JSPException(ioexc.toString());
		}
	}

	private void compileJavaFile() throws JSPException {
		String[]				javacArgs;
		ByteArrayOutputStream	javacOut;

		javacArgs = new String[]{
					"-d",
					classFilesDir.getPath(),
					javaFile.getPath()
						};
		javacOut = new ByteArrayOutputStream();
		sun.tools.javac.Main javaCompiler = new sun.tools.javac.Main(new PrintStream(javacOut, true), "javac");
		if(!javaCompiler.compile(javacArgs)) {
			throw new JSPException(javacOut.toString());
		}
	}

	private static String cEncode(String val) {
		StringBuffer	res;
		int				i;
		char			c;

		res = new StringBuffer(val.length());
		for(i = 0; i < val.length(); i++) {
			c = val.charAt(i);
			switch(c) {
			case '\n':
				res.append("\\n");
				break;
			case '\r':
				res.append("\\r");
				break;
			case '\'':
				res.append("\\\'");
				break;
			case '\"':
				res.append("\\\"");
				break;
			default:
				res.append((char) c);
				break;
			}
		}
		return res.toString();
	}

	public static String getClassNameForJspFile(File jspFile) throws JSPException {
		StringTokenizer		toker;
		StringBuffer		buf;
		String				s;
		char				c;
		int					i;

		try {
			s = jspFile.getCanonicalPath();
		} catch(IOException ioexc) {
			throw new JSPException("Could not determine full name for JSP file " + jspFile);
		}
		toker = new StringTokenizer(s, String.valueOf(File.separatorChar));
		buf = new StringBuffer(s.length());
		while(toker.hasMoreTokens()) {
			s = toker.nextToken();
			if(buf.length() > 0) {
				buf.append((char) '.');
			}
			if(!toker.hasMoreTokens()) {
				buf.append((char) '_');
			}	
			for(i = 0; i < s.length(); i++) {
				c = s.charAt(i);
				if(Character.isJavaIdentifierPart(c)) {
					buf.append((char) c);
				} else {
					buf.append((char) '_').append(Integer.toHexString(c));
				}
			}
		}

		return buf.toString();
	}

	public static File getFileForClass(File classFilesDir, String className, String suffix) {
		return new File(classFilesDir, className.replace('.', File.separatorChar) + suffix);
	}

	public static void main(String args[]) {
		File			javaFiles = new File("/Bas/Incoming/");
		Vector			errs = new Vector();
		try {
			File f = new File("test.jsp");
			String className = getClassNameForJspFile(f);
			JSPCompiler compiler = new JSPCompiler(javaFiles, f, className);

			compiler.compile();
			System.out.println("compiled class = " + className);
			JSPClassLoader clldr = new JSPClassLoader(javaFiles);
			Class c = clldr.loadClass(className);
			if(c != null) {
				System.out.println("class loaded: " + c);
				System.out.println("superclass = " + c.getSuperclass());
				Object o = c.newInstance();
				System.out.println("running it!");
				((javax.servlet.Servlet) o).service(null, null);
				System.out.println("test0: " + (c.getClassLoader() == clldr));

				System.out.println("loading it again");
				clldr = new JSPClassLoader(javaFiles);
				Class c2 = clldr.loadClass(className);
				if(c == c2)
					System.out.println("jsp classes cached?");
				else
					System.out.println("jsp classes not cached!");
				System.out.println("test1: " + (c2.getClassLoader() == clldr));
				System.out.println("test2: " + (c.getClassLoader() == clldr));
				
			} else {
				System.out.println("class not loaded");
			}
		} catch(Exception exc) {
			exc.printStackTrace();
		}
	}
}
