/*
	GnuJSP - a free JSP implementation
	Copyright (C) 1998, Vincent Partington <vinny@nmg.nl>
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package	nl.nmg.jsp;
import	java.io.File;
import	java.io.IOException;
import	java.io.PrintWriter;
import	java.lang.reflect.Field;
import	java.lang.reflect.InvocationTargetException;
import	java.lang.reflect.Method;
import	java.util.Enumeration;
import	java.util.Hashtable;
import	java.util.Vector;
import	javax.servlet.*;
import	javax.servlet.http.*;

public class JSPServlet extends HttpServlet {
	private File			classFilesDir;
	private JSPClassLoader	jspClassLoader;
	private Hashtable		jspServlets;

	public void init(ServletConfig config) throws ServletException {
		super.init(config);

		String s = config.getInitParameter("classfiles");
		if(s == null)
			throw new ServletException("no classfiles init parameter specified");
		classFilesDir = new File(s);
		jspClassLoader = new JSPClassLoader(classFilesDir);
		jspServlets = new Hashtable();
	}

	public void service(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		File			jspFile;
		String			className;
		Servlet			srv;
		Class			c;
		Method			meth;
		Boolean			depCheckRes;
		JSPCompiler		jspCompiler;

		try {
			// determine jsp file and class name
			if(request.getPathTranslated() == null) {
				throw new JSPException("No JSP page specified");
			}
			jspFile = new File(request.getPathTranslated());
			className = JSPCompiler.getClassNameForJspFile(jspFile);

			// load jsp servlet from cache or class file
			srv = (Servlet) jspServlets.get(className);
			if(srv == null) {
				srv = loadServlet(className);
			}
	
			// unload jsp servlet if dependencies demand it
			if(srv != null) {
				c = srv.getClass();
				try {
					// invoke dependency checking method
					meth = c.getMethod("__checkDependencies", new Class[0]);
					depCheckRes = (Boolean) meth.invoke(srv, new Object[0]);
					if(depCheckRes.booleanValue()) {
						unloadServlet(srv);
						refreshClassLoader(c);
						srv = null;
					}
				} catch(IllegalAccessException iaexc) {
					;
				} catch(InvocationTargetException itexc) {
					;
				} catch(NoSuchMethodException nsmexc) {
					;
				} catch(SecurityException secexc) {
					;
				} finally {
					c = null;
				}
			}

			// compile & load jsp servlet if we haven't loaded it yet
			if(srv == null) {
				jspCompiler = new JSPCompiler(classFilesDir, jspFile, className);
				jspCompiler.compile();
	
				srv = loadServlet(className);
				if(srv == null) {
					throw new JSPException("Could not load jsp servlet class " + className);
				}
			}
		} catch(JSPException jspexc) {
			response.setContentType("text/html");
			PrintWriter out = response.getWriter();
			out.println("<H3>An error occurred</H3><PRE>");
			out.println(jspexc.getMessage());
			out.println("</PRE>");
			out.close();
			return;
		}

		srv.service(request, response);
	}

	private Servlet loadServlet(String className) throws JSPException {
		Class		c;
		Field		fld;
		Servlet		srv;

		// load servlet class
		try {
			c = jspClassLoader.loadClass(className);
			if(c == null)
				return null;
		} catch(ClassNotFoundException cnfexc) {
			return null;
		}

		// check compiler version dependency
		try {
			fld = c.getField("__compilerVersionNr");
			if(fld.getInt(null) != JSPCompiler.COMPILER_VERSION_NR) {
				refreshClassLoader(c);
				return null;
			}
		} catch(IllegalAccessException fiaexc) {
			throw new JSPException("Jsp servlet class " + className + " does not have a visible __compilerVersionNr field");
		} catch(NoSuchFieldException nsfexc) {
			throw new JSPException("Jsp servlet class " + className + " does not have a __compilerVersionNr field");
		} catch(SecurityException secexc) {
			throw new JSPException("Jsp servlet class " + className + " has a security problem with its __compilerVersionNr field");
		}

		// instantiate servlet
		try {
			srv = (Servlet) c.newInstance();
		} catch(IllegalAccessException iaexc) {
			throw new JSPException("Could not instantiate jsp servlet class " + className + " because of an illegal access exception");
		} catch(InstantiationException iexc) {
			throw new JSPException("Could not instantiate jsp servlet class " + className + " because it is an interface or an abstract class");
		}

		// init servlet
		try {
			srv.init(new ServletConfig() {
				public ServletContext getServletContext() {
					return JSPServlet.this.getServletContext();
				}
				public String getInitParameter(String name) {
					return null;
				}
				public Enumeration getInitParameterNames() {
					return (new Vector()).elements();
				}
			});
		} catch(Exception exc) {}

		jspServlets.put(className, srv);

		return srv;
	}

	private void unloadServlet(Servlet srv) {
		try {
			srv.destroy();
		} catch(Exception exc) { }

		jspServlets.remove(srv.getClass().getName());
	}

	private void refreshClassLoader(Class c) {
		if(c.getClassLoader() == jspClassLoader) {
			jspClassLoader = new JSPClassLoader(classFilesDir);
		}
	}

	public void destroy() {
		super.destroy();

		Enumeration		e;
		Servlet			srv;

		e = jspServlets.elements();
		while(e.hasMoreElements()) {
			srv = (Servlet) e.nextElement();
			unloadServlet(srv);
		}
	}
}
